@@ -8,8 +8,7 @@ from contextlib import AsyncExitStack
 from contextlib import ExitStack
 from contextlib import contextmanager
 from contextlib import suppress
-from os.path import exists
-from os.path import join
+from pathlib import Path
 from pprint import pformat
 from typing import Any
 from typing import Awaitable
@@ -31,6 +30,7 @@ from dipdup.config import BigMapIndexConfig
 from dipdup.config import ContractConfig
 from dipdup.config import DipDupConfig
 from dipdup.config import EventHookConfig
+from dipdup.config import EventIndexConfig
 from dipdup.config import HandlerConfig
 from dipdup.config import HeadIndexConfig
 from dipdup.config import HookConfig
@@ -262,12 +262,13 @@ class DipDupContext:
     async def _spawn_index(self, name: str, state: Optional[Index] = None) -> None:
         # NOTE: Avoiding circular import
         from dipdup.index import BigMapIndex
+        from dipdup.index import EventIndex
         from dipdup.index import HeadIndex
         from dipdup.index import OperationIndex
         from dipdup.index import TokenTransferIndex
 
         index_config = cast(ResolvedIndexConfigT, self.config.get_index(name))
-        index: Union[OperationIndex, BigMapIndex, HeadIndex, TokenTransferIndex]
+        index: OperationIndex | BigMapIndex | HeadIndex | TokenTransferIndex | EventIndex
 
         datasource_name = cast(TzktDatasourceConfig, index_config.datasource).name
         datasource = self.get_tzkt_datasource(datasource_name)
@@ -280,6 +281,8 @@ class DipDupContext:
             index = HeadIndex(self, index_config, datasource)
         elif isinstance(index_config, TokenTransferIndexConfig):
             index = TokenTransferIndex(self, index_config, datasource)
+        elif isinstance(index_config, EventIndexConfig):
+            index = EventIndex(self, index_config, datasource)
         else:
             raise NotImplementedError
 
@@ -589,13 +592,9 @@ class CallbackManager:
 
     async def execute_sql(self, ctx: 'DipDupContext', name: str) -> None:
         """Execute SQL included with project"""
-        if not isinstance(ctx.config.database, PostgresDatabaseConfig):
-            self._logger.warning('Skipping SQL script `%s`: not supported on SQLite', name)
-            return
-
         subpackages = name.split('.')
-        sql_path = join(ctx.config.package_path, 'sql', *subpackages)
-        if not exists(sql_path):
+        sql_path = Path(ctx.config.package_path, 'sql', *subpackages)
+        if not sql_path.exists():
             raise InitializationRequiredError(f'Missing SQL directory for hook `{name}`')
 
         # NOTE: SQL scripts are not wrapped in transaction
