@@ -4,32 +4,112 @@ from functools import lru_cache
 from itertools import groupby
 from typing import Any
 from typing import Dict
+from typing import Hashable
 from typing import Iterable
 from typing import List
+from typing import Literal
 from typing import Optional
 from typing import Tuple
 from typing import Type
+from typing import TypeVar
 from typing import Union
+from typing import cast
 
+from pydantic import BaseModel
 from pydantic import Extra
-from pydantic.error_wrappers import ValidationError
+from pydantic.dataclasses import dataclass
 from typing_extensions import get_args
 from typing_extensions import get_origin
 
-from dipdup.exceptions import InvalidDataError
+from dipdup.datasources.subscription import Subscription
 from dipdup.models import OperationData
 from dipdup.models import StorageType
+from dipdup.utils.codegen import parse_object
 
 IntrospectionError = (KeyError, IndexError, AttributeError)
 
+T = TypeVar('T', Hashable, Type[BaseModel])
 
-def extract_root_outer_type(storage_type: Type) -> Type:
+
+@dataclass(frozen=True)
+class HeadSubscription(Subscription):
+    type: Literal['head'] = 'head'
+    method: Literal['SubscribeToHead'] = 'SubscribeToHead'
+
+    def get_request(self) -> Any:
+        return []
+
+
+@dataclass(frozen=True)
+class OriginationSubscription(Subscription):
+    type: Literal['origination'] = 'origination'
+    method: Literal['SubscribeToOperations'] = 'SubscribeToOperations'
+
+    def get_request(self) -> Any:
+        return [{'types': 'origination'}]
+
+
+@dataclass(frozen=True)
+class TransactionSubscription(Subscription):
+    type: Literal['transaction'] = 'transaction'
+    method: Literal['SubscribeToOperations'] = 'SubscribeToOperations'
+    address: Optional[str] = None
+
+    def get_request(self) -> Any:
+        request = [{'types': 'transaction'}]
+        if self.address:
+            request[0]['address'] = self.address
+        return request
+
+
+# TODO: Add `ptr` and `tags` filters?
+@dataclass(frozen=True)
+class BigMapSubscription(Subscription):
+    type: Literal['big_map'] = 'big_map'
+    method: Literal['SubscribeToBigMaps'] = 'SubscribeToBigMaps'
+    address: Optional[str] = None
+    path: Optional[str] = None
+
+    def get_request(self) -> Any:
+        if self.address and self.path:
+            return [{'address': self.address, 'paths': [self.path]}]
+        elif not self.address and not self.path:
+            return [{}]
+        else:
+            raise RuntimeError
+
+
+@dataclass(frozen=True)
+class TokenTransferSubscription(Subscription):
+    type: Literal['token_transfer'] = 'token_transfer'
+    method: Literal['SubscribeToTokenTransfers'] = 'SubscribeToTokenTransfers'
+
+    # NOTE: No filters applied
+    def get_request(self) -> Any:
+        return [{}]
+
+
+@dataclass(frozen=True)
+class EventSubscription(Subscription):
+    type: Literal['event'] = 'event'
+    method: Literal['SubscribeToEvents'] = 'SubscribeToEvents'
+    address: Optional[str] = None
+
+    def get_request(self) -> Any:
+        if self.address:
+            return [{'address': self.address}]
+
+        return [{}]
+
+
+def extract_root_outer_type(storage_type: Type[BaseModel]) -> T:
     """Extract Pydantic __root__ type"""
     root_field = storage_type.__fields__['__root__']
     if root_field.allow_none:
-        return typing.Optional[root_field.type_]  # type: ignore
-    else:
-        return root_field.outer_type_
+        # NOTE: Optional is a magic _SpecialForm
+        return cast(Type[BaseModel], typing.Optional[root_field.type_])
+
+    return root_field.outer_type_
 
 
 @lru_cache(None)
@@ -42,7 +122,7 @@ def is_array_type(storage_type: Type) -> bool:
     # NOTE: Pydantic model with __root__ field subclassing List
     with suppress(*IntrospectionError):
         root_type = extract_root_outer_type(storage_type)
-        return is_array_type(root_type)  # type: ignore
+        return is_array_type(root_type)
 
     # NOTE: Something else
     return False
@@ -57,7 +137,7 @@ def get_list_elt_type(list_type: Type[Any]) -> Type[Any]:
 
     # NOTE: Pydantic model with __root__ field subclassing List
     root_type = extract_root_outer_type(list_type)
-    return get_list_elt_type(root_type)  # type: ignore
+    return get_list_elt_type(root_type)
 
 
 @lru_cache(None)
@@ -70,7 +150,7 @@ def get_dict_value_type(dict_type: Type[Any], key: Optional[str] = None) -> Type
     # NOTE: Pydantic model with __root__ field subclassing Dict
     with suppress(*IntrospectionError):
         root_type = extract_root_outer_type(dict_type)
-        return get_dict_value_type(root_type, key)  # type: ignore
+        return get_dict_value_type(root_type, key)
 
     if key is None:
         raise KeyError('Field name or alias is required for object introspection')
@@ -81,7 +161,7 @@ def get_dict_value_type(dict_type: Type[Any], key: Optional[str] = None) -> Type
         if key in (field.name, field.alias):
             # NOTE: Pydantic does not preserve outer_type_ for Optional
             if field.allow_none:
-                return typing.Optional[field.type_]  # type: ignore
+                return cast(Type[Any], typing.Optional[field.type_])
             else:
                 return field.outer_type_
 
@@ -97,7 +177,7 @@ def unwrap_union_type(union_type: Type) -> Tuple[bool, Tuple[Type, ...]]:
 
     with suppress(*IntrospectionError):
         root_type = extract_root_outer_type(union_type)
-        return unwrap_union_type(root_type)  # type: ignore
+        return unwrap_union_type(root_type)
 
     return False, ()
 
@@ -135,10 +215,10 @@ def _apply_bigmap_diffs(
         return dict_storage
 
 
-def _process_storage(storage: Any, storage_type: Type[Any], bigmap_diffs: Dict[int, Iterable[Dict[str, Any]]]) -> Any:
+def _process_storage(storage: Any, storage_type: T, bigmap_diffs: Dict[int, Iterable[Dict[str, Any]]]) -> Any:
     """Replace bigmap pointers with actual data from diffs"""
     # Check if Union or Optional (== Union[Any, NoneType])
-    is_union, arg_types = unwrap_union_type(storage_type)  # type: ignore
+    is_union, arg_types = unwrap_union_type(storage_type)
     if is_union:
         # NOTE: We have no way but trying every possible branch until first success
         for arg_type in arg_types:
@@ -147,12 +227,12 @@ def _process_storage(storage: Any, storage_type: Type[Any], bigmap_diffs: Dict[i
 
     # NOTE: Bigmap pointer, apply diffs
     if isinstance(storage, int) and type(storage) != storage_type:
-        is_array = is_array_type(storage_type)  # type: ignore
+        is_array = is_array_type(storage_type)
         storage = _apply_bigmap_diffs(storage, bigmap_diffs, is_array)
 
     # NOTE: List, process recursively
     elif isinstance(storage, list):
-        elt_type = get_list_elt_type(storage_type)  # type: ignore
+        elt_type = get_list_elt_type(storage_type)
         for i, _ in enumerate(storage):
             storage[i] = _process_storage(storage[i], elt_type, bigmap_diffs)
 
@@ -163,7 +243,7 @@ def _process_storage(storage: Any, storage_type: Type[Any], bigmap_diffs: Dict[i
 
         for key, value in storage.items():
             try:
-                value_type = get_dict_value_type(storage_type, key)  # type: ignore
+                value_type = get_dict_value_type(storage_type, key)
                 storage[key] = _process_storage(value, value_type, bigmap_diffs)
             except IntrospectionError:
                 if not ignore:
@@ -184,12 +264,4 @@ def deserialize_storage(operation_data: OperationData, storage_type: Type[Storag
         storage_type=storage_type,
         bigmap_diffs=bigmap_diffs,
     )
-
-    try:
-        return storage_type.parse_obj(operation_data.storage)
-    except ValidationError as e:
-        raise InvalidDataError(
-            type_cls=storage_type,
-            data=operation_data.storage,
-            parsed_object=operation_data,
-        ) from e
+    return parse_object(storage_type, operation_data.storage)
