@@ -1,11 +1,20 @@
 from enum import Enum
 
 
+class LoggingValues(Enum):
+    """Enum for `logging` field values."""
+
+    default = 'default'
+    quiet = 'quiet'
+    verbose = 'verbose'
+
+
 class MessageType(Enum):
     operation = 'operation'
     big_map = 'big_map'
     head = 'head'
     token_transfer = 'token_transfer'
+    event = 'event'
 
 
 class IndexType(Enum):
@@ -13,6 +22,7 @@ class IndexType(Enum):
     big_map = 'big_map'
     head = 'head'
     token_transfer = 'token_transfer'
+    event = 'event'
 
 
 class OperationType(Enum):
