@@ -2,7 +2,7 @@ import tempfile
 import textwrap
 from dataclasses import dataclass
 from dataclasses import field
-from os.path import join
+from pathlib import Path
 from tempfile import NamedTemporaryFile
 from typing import Any
 from typing import Dict
@@ -16,7 +16,7 @@ from tortoise.models import Model
 from dipdup import spec_version_mapping
 from dipdup.enums import ReindexingReason
 
-_tab = ('_' * 80) + '\n\n'
+tab = ('_' * 80) + '\n\n'
 
 
 def unindent(text: str) -> str:
@@ -35,14 +35,12 @@ def save_crashdump(error: Exception) -> str:
     import sentry_sdk.serializer
     import sentry_sdk.utils
 
-    from dipdup.utils import mkdir_p
-
     exc_info = sentry_sdk.utils.exc_info_from_error(error)
     event, _ = sentry_sdk.utils.event_from_exception(exc_info)
     event = sentry_sdk.serializer.serialize(event)
 
-    tmp_dir = join(tempfile.gettempdir(), 'dipdup', 'crashdumps')
-    mkdir_p(tmp_dir)
+    tmp_dir = Path(tempfile.gettempdir()) / 'dipdup' / 'crashdumps'
+    tmp_dir.mkdir(parents=True, exist_ok=True)
 
     crashdump_file = NamedTemporaryFile(
         mode='ab',
@@ -91,7 +89,7 @@ class DipDupError(Exception):
         return unindent(self._help())
 
     def format(self) -> str:
-        return _tab + self.help() + '\n'
+        return tab + self.help() + '\n'
 
 
 @dataclass(repr=False)
@@ -177,7 +175,7 @@ class MigrationRequiredError(DipDupError):
             ],
             headers=['', 'spec_version', 'DipDup version'],
         )
-        reindex = '\n\n' + _tab + ReindexingRequiredError(ReindexingReason.migration).help() if self.reindex else ''
+        reindex = '\n\n' + tab + ReindexingRequiredError(ReindexingReason.migration).help() if self.reindex else ''
         return f"""
             Project migration required!
 
@@ -285,7 +283,7 @@ class ContractAlreadyExistsError(DipDupError):
 
 @dataclass(repr=False)
 class IndexAlreadyExistsError(DipDupError):
-    """Attemp to add an index with an alias already in use"""
+    """Attempt to add an index with an alias already in use"""
 
     ctx: Any
     name: str
@@ -310,23 +308,19 @@ class IndexAlreadyExistsError(DipDupError):
 class InvalidDataError(DipDupError):
     """Failed to validate datasource message against generated type class"""
 
-    type_cls: Type[Any]
+    msg: str
+    type_: Type[Any]
     data: Any
-    parsed_object: Any
 
     def _help(self) -> str:
 
         return f"""
             Failed to validate datasource message against generated type class.
 
-            Expected type:
-            `{self.type_cls.__name__}`
+              {self.msg}
 
-            Invalid data:
-            {self.data}
-
-            Parsed object:
-            {self.parsed_object}
+            Type class: `{self.type_.__name__}`
+            Data: `{self.data}`
         """
 
 
@@ -388,6 +382,24 @@ class HasuraError(DipDupError):
             If it's `400 Bad Request`, check out Hasura logs for more information.
 
             See https://docs.dipdup.io/graphql/
-            See https://docs.dipdup.io/config/hasura.html
-            See https://docs.dipdup.io/cli-reference.html#dipdup-hasura-configure
+            See https://docs.dipdup.io/config/hasura
+            See https://docs.dipdup.io/cli-reference#dipdup-hasura-configure
+        """
+
+
+@dataclass(repr=False)
+class FeatureAvailabilityError(DipDupError):
+    """Requested feature is not supported in the current environment"""
+
+    feature: str
+    reason: str
+
+    def _help(self) -> str:
+        return f"""
+            Feature `{self.feature}` is not available in the current environment.
+
+            {self.reason}
+
+            See https://docs.dipdup.io/installation
+            See https://docs.dipdup.io/advanced/docker
         """
