@@ -2,8 +2,9 @@ import logging
 from abc import abstractmethod
 from collections import defaultdict
 from collections import deque
-from collections import namedtuple
 from contextlib import ExitStack
+from contextlib import suppress
+from copy import copy
 from datetime import datetime
 from typing import DefaultDict
 from typing import Deque
@@ -19,12 +20,14 @@ from typing import Union
 from typing import cast
 
 from pydantic.dataclasses import dataclass
-from pydantic.error_wrappers import ValidationError
 
 import dipdup.models as models
 from dipdup.config import BigMapHandlerConfig
 from dipdup.config import BigMapIndexConfig
 from dipdup.config import ContractConfig
+from dipdup.config import EventHandlerConfig
+from dipdup.config import EventHandlerConfigT
+from dipdup.config import EventIndexConfig
 from dipdup.config import HeadHandlerConfig
 from dipdup.config import HeadIndexConfig
 from dipdup.config import OperationHandlerConfig
@@ -37,8 +40,11 @@ from dipdup.config import ResolvedIndexConfigT
 from dipdup.config import SkipHistory
 from dipdup.config import TokenTransferHandlerConfig
 from dipdup.config import TokenTransferIndexConfig
+from dipdup.config import UnknownEventHandlerConfig
 from dipdup.context import DipDupContext
+from dipdup.context import rolled_back_indexes
 from dipdup.datasources.tzkt.datasource import BigMapFetcher
+from dipdup.datasources.tzkt.datasource import EventFetcher
 from dipdup.datasources.tzkt.datasource import OperationFetcher
 from dipdup.datasources.tzkt.datasource import TokenTransferFetcher
 from dipdup.datasources.tzkt.datasource import TzktDatasource
@@ -47,19 +53,21 @@ from dipdup.enums import MessageType
 from dipdup.exceptions import ConfigInitializationException
 from dipdup.exceptions import ConfigurationError
 from dipdup.exceptions import InvalidDataError
-from dipdup.exceptions import ReindexingReason
 from dipdup.models import BigMapAction
 from dipdup.models import BigMapData
 from dipdup.models import BigMapDiff
+from dipdup.models import Event
+from dipdup.models import EventData
 from dipdup.models import HeadBlockData
 from dipdup.models import IndexStatus
 from dipdup.models import OperationData
 from dipdup.models import Origination
 from dipdup.models import TokenTransferData
 from dipdup.models import Transaction
+from dipdup.models import UnknownEvent
 from dipdup.prometheus import Metrics
 from dipdup.utils import FormattedLogger
-from dipdup.utils.database import in_global_transaction
+from dipdup.utils.codegen import parse_object
 
 _logger = logging.getLogger(__name__)
 
@@ -77,13 +85,14 @@ class OperationSubgroup:
         return hash(f'{self.hash}:{self.counter}')
 
 
-# NOTE: Message queue of OperationIndex
-SingleLevelRollback = namedtuple('SingleLevelRollback', ('from_level'))
 Operations = Tuple[OperationData, ...]
-OperationQueueItemT = Union[Tuple[OperationSubgroup, ...], SingleLevelRollback]
 OperationHandlerArgumentT = Optional[Union[Transaction, Origination, OperationData]]
 MatchedOperationsT = Tuple[OperationSubgroup, OperationHandlerConfig, Deque[OperationHandlerArgumentT]]
 MatchedBigMapsT = Tuple[BigMapHandlerConfig, BigMapDiff]
+MatchedEventsT = Union[
+    Tuple[EventHandlerConfig, Event],
+    Tuple[UnknownEventHandlerConfig, UnknownEvent],
+]
 
 
 def extract_operation_subgroups(
@@ -176,6 +185,17 @@ class Index:
     def realtime(self) -> bool:
         return self.state.status == IndexStatus.REALTIME and not self._queue
 
+    def get_sync_level(self) -> int:
+        """Get level index needs to be synchronized to depending on its subscription status"""
+        sync_levels = {self.datasource.get_sync_level(s) for s in self._config.subscriptions}
+        if not sync_levels:
+            raise RuntimeError('Index has no subscriptions')
+        if None in sync_levels:
+            raise RuntimeError('Call `set_sync_level` before starting IndexDispatcher')
+        # NOTE: Multiple sync levels means index with new subscriptions was added in runtime.
+        # NOTE: Choose the highest level; outdated realtime messages will be dropped from the queue anyway.
+        return max(cast(Set[int], sync_levels))
+
     async def initialize_state(self, state: Optional[models.Index] = None) -> None:
         if self._state:
             raise RuntimeError('Index state is already initialized')
@@ -200,7 +220,11 @@ class Index:
             },
         )
 
-    async def process(self) -> None:
+    async def process(self) -> bool:
+        if self.name in rolled_back_indexes:
+            await self.state.refresh_from_db(('level',))
+            rolled_back_indexes.remove(self.name)
+
         if not isinstance(self._config, HeadIndexConfig) and self._config.last_level:
             head_level = self._config.last_level
             with ExitStack() as stack:
@@ -209,13 +233,8 @@ class Index:
                 await self._synchronize(head_level)
             await self.state.update_status(IndexStatus.ONESHOT, head_level)
 
-        sync_levels = {self.datasource.get_sync_level(s) for s in self._config.subscriptions}
-        sync_level = sync_levels.pop()
-        if sync_level is None:
-            raise RuntimeError('Call `set_sync_level` before starting IndexDispatcher')
-        if sync_levels:
-            raise RuntimeError(f'Multiple sync levels: {sync_level}, {sync_levels}')
         index_level = self.state.level
+        sync_level = self.get_sync_level()
 
         if index_level < sync_level:
             self._logger.info('Index is behind datasource level, syncing: %s -> %s', index_level, sync_level)
@@ -231,6 +250,9 @@ class Index:
                 if Metrics.enabled:
                     stack.enter_context(Metrics.measure_total_realtime_duration())
                 await self._process_queue()
+        else:
+            return False
+        return True
 
     @abstractmethod
     async def _synchronize(self, head_level: int) -> None:
@@ -264,7 +286,7 @@ class Index:
 
     def _extract_level(
         self,
-        message: Union[Tuple[OperationData, ...], Tuple[BigMapData, ...], Tuple[TokenTransferData, ...]],
+        message: tuple[OperationData | BigMapData | TokenTransferData | EventData, ...],
     ) -> int:
         batch_levels = {item.level for item in message}
         if len(batch_levels) != 1:
@@ -278,37 +300,14 @@ class OperationIndex(Index):
 
     def __init__(self, ctx: DipDupContext, config: OperationIndexConfig, datasource: TzktDatasource) -> None:
         super().__init__(ctx, config, datasource)
-        self._queue: Deque[OperationQueueItemT] = deque()
+        self._queue: Deque[Tuple[OperationSubgroup, ...]] = deque()
         self._contract_hashes: Dict[str, Tuple[int, int]] = {}
-        self._next_head_level: Optional[int] = None
-        self._head_hashes: Dict[str, bool] = {}
-        self._migration_originations: Optional[Dict[str, OperationData]] = None
 
     def push_operations(self, operation_subgroups: Tuple[OperationSubgroup, ...]) -> None:
         self._queue.append(operation_subgroups)
         if Metrics.enabled:
             Metrics.set_levels_to_realtime(self._config.name, len(self._queue))
 
-    def push_rollback(self, from_level: int) -> None:
-        self._queue.append(SingleLevelRollback(from_level))
-
-    async def _single_level_rollback(self, from_level: int) -> None:
-        """Ensure the next arrived block has all operations of the previous one. But it could also contain additional operations we need to process.
-
-        Called by IndexDispatcher when index datasource receive a single level rollback.
-        """
-        if self._next_head_level:
-            raise RuntimeError('Index is already in a single-level rollback state')
-
-        index_level = cast(int, self.state.level)
-        if index_level < from_level:
-            self._logger.info('Index level is lower than new head level, ignoring: %s < %s', index_level, from_level)
-        elif index_level == from_level:
-            self._logger.info('Single level rollback, next block will be processed partially')
-            self._next_head_level = from_level
-        else:
-            raise RuntimeError(f'Index level is higher than new head level: {index_level} > {from_level}')
-
     async def _process_queue(self) -> None:
         """Process WebSocket queue"""
         self._logger.debug('Processing %s realtime messages from queue', len(self._queue))
@@ -323,37 +322,30 @@ class OperationIndex(Index):
             if Metrics.enabled:
                 Metrics.set_levels_to_realtime(self._config.name, messages_left)
 
-            if isinstance(message, SingleLevelRollback):
-                # NOTE: To match <= condition, variable is not used anywhere else
-                message_level = message.from_level + 1
-            else:
-                message_level = message[0].operations[0].level
+            message_level = message[0].operations[0].level
 
             if message_level <= self.state.level:
                 self._logger.debug('Skipping outdated message: %s <= %s', message_level, self.state.level)
                 continue
 
-            if isinstance(message, SingleLevelRollback):
-                await self._single_level_rollback(message.from_level)
-            else:
-                with ExitStack() as stack:
-                    if Metrics.enabled:
-                        stack.enter_context(Metrics.measure_level_realtime_duration())
-                    await self._process_level_operations(message)
+            with ExitStack() as stack:
+                if Metrics.enabled:
+                    stack.enter_context(Metrics.measure_level_realtime_duration())
+                await self._process_level_operations(message, message_level)
 
         else:
             if Metrics.enabled:
                 Metrics.set_levels_to_realtime(self._config.name, 0)
 
-    async def _synchronize(self, head_level: int) -> None:
+    async def _synchronize(self, sync_level: int) -> None:
         """Fetch operations via Fetcher and pass to message callback"""
-        index_level = await self._enter_sync_state(head_level)
+        index_level = await self._enter_sync_state(sync_level)
         if index_level is None:
             return
 
         first_level = index_level + 1
 
-        self._logger.info('Fetching operations from level %s to %s', first_level, head_level)
+        self._logger.info('Fetching operations from level %s to %s', first_level, sync_level)
         transaction_addresses = await self._get_transaction_addresses()
         origination_addresses = await self._get_origination_addresses()
 
@@ -368,7 +360,7 @@ class OperationIndex(Index):
         fetcher = OperationFetcher(
             datasource=self._datasource,
             first_level=first_level,
-            last_level=head_level,
+            last_level=sync_level,
             transaction_addresses=transaction_addresses,
             origination_addresses=origination_addresses,
             migration_originations=migration_originations,
@@ -376,7 +368,7 @@ class OperationIndex(Index):
 
         async for level, operations in fetcher.fetch_operations_by_level():
             if Metrics.enabled:
-                Metrics.set_levels_to_sync(self._config.name, head_level - level)
+                Metrics.set_levels_to_sync(self._config.name, sync_level - level)
 
             operation_subgroups = tuple(
                 extract_operation_subgroups(
@@ -390,68 +382,27 @@ class OperationIndex(Index):
                 with ExitStack() as stack:
                     if Metrics.enabled:
                         stack.enter_context(Metrics.measure_level_sync_duration())
-                    await self._process_level_operations(operation_subgroups)
+                    await self._process_level_operations(operation_subgroups, sync_level)
 
-        await self._exit_sync_state(head_level)
+        await self._exit_sync_state(sync_level)
 
-    async def _process_level_operations(self, operation_subgroups: Tuple[OperationSubgroup, ...]) -> None:
+    async def _process_level_operations(
+        self,
+        operation_subgroups: Tuple[OperationSubgroup, ...],
+        sync_level: int,
+    ) -> None:
         if not operation_subgroups:
-            raise RuntimeError('No operations to process')
+            return
 
         batch_level = operation_subgroups[0].operations[0].level
         index_level = self.state.level
-        if batch_level < index_level:
-            raise RuntimeError(f'Batch level is lower than index level: {batch_level} < {index_level}')
-
-        # NOTE: Single level rollback was triggered
-        if head_level := self._next_head_level:
-            if head_level != index_level:
-                raise RuntimeError(f'New head level is not equal to index level: {head_level} != {index_level}')
-
-            self._logger.info('Rolling back to the previous level, verifying processed operations')
-            rollback_hook_called = False
-            old_head_hashes = set(self._head_hashes)
-            old_head_matched_hashes = {k for k, v in self._head_hashes.items() if v}
-            new_head_hashes = {s.hash for s in operation_subgroups}
-            unprocessed_hashes = new_head_hashes - old_head_hashes
-            # NOTE: We can ignore subgroups that don't match any handlers
-            missing_hashes = old_head_matched_hashes - new_head_hashes
-
-            self._logger.info('Comparing hashes: %s new, %s missing', len(unprocessed_hashes), len(missing_hashes))
-            if missing_hashes:
-                rollback_hook_called = True
-                self._logger.info('Some operations were backtracked, calling rollback hook')
-                if self._ctx.config.per_index_rollback:
-                    hook_name = 'on_index_rollback'
-                    await self._ctx.fire_hook(
-                        hook_name,
-                        index=self,
-                        from_level=head_level,
-                        to_level=head_level - 1,
-                    )
-                else:
-                    hook_name = 'on_rollback'
-                    await self._ctx.fire_hook(
-                        hook_name,
-                        datasource=self.datasource,
-                        from_level=head_level + 1,
-                        to_level=head_level,
-                    )
-
-            self._next_head_level = None
-            self._head_hashes.clear()
-
-            if rollback_hook_called:
-                return
-
-            operation_subgroups = tuple(filter(lambda subgroup: subgroup.hash in unprocessed_hashes, operation_subgroups))
+        if batch_level <= index_level:
+            raise RuntimeError(f'Batch level is lower than index level: {batch_level} <= {index_level}')
 
         self._logger.debug('Processing %s operation subgroups of level %s', len(operation_subgroups), batch_level)
         matched_handlers: Deque[MatchedOperationsT] = deque()
         for operation_subgroup in operation_subgroups:
-            subgroup_matched_handlers = await self._match_operation_subgroup(operation_subgroup)
-            matched_handlers += subgroup_matched_handlers
-            self._head_hashes[operation_subgroup.hash] = bool(subgroup_matched_handlers)
+            matched_handlers += await self._match_operation_subgroup(operation_subgroup)
 
         if Metrics.enabled:
             Metrics.set_index_handlers_matched(len(matched_handlers))
@@ -461,7 +412,7 @@ class OperationIndex(Index):
             await self.state.update_status(level=batch_level)
             return
 
-        async with in_global_transaction():
+        async with self._ctx._transactions.in_transaction(batch_level, sync_level, self.name):
             for operation_subgroup, handler_config, args in matched_handlers:
                 await self._call_matched_handler(handler_config, operation_subgroup, args)
             await self.state.update_status(level=batch_level)
@@ -564,11 +515,8 @@ class OperationIndex(Index):
                     args.append(operation_data)
                     continue
 
-                parameter_type = pattern_config.parameter_type_cls
-                try:
-                    parameter = parameter_type.parse_obj(operation_data.parameter_json) if parameter_type else None
-                except ValidationError as e:
-                    raise InvalidDataError(parameter_type, operation_data.parameter_json, operation_data) from e
+                type_ = pattern_config.parameter_type_cls
+                parameter = parse_object(type_, operation_data.parameter_json) if type_ else None
 
                 storage_type = pattern_config.storage_type_cls
                 storage = deserialize_storage(operation_data, storage_type)
@@ -601,7 +549,7 @@ class OperationIndex(Index):
         if not handler_config.parent:
             raise ConfigInitializationException
 
-        await self._ctx.fire_handler(
+        await self._ctx._fire_handler(
             handler_config.callback,
             handler_config.parent.name,
             self.datasource,
@@ -676,11 +624,11 @@ class BigMapIndex(Index):
             with ExitStack() as stack:
                 if Metrics.enabled:
                     stack.enter_context(Metrics.measure_level_realtime_duration())
-                await self._process_level_big_maps(big_maps)
+                await self._process_level_big_maps(big_maps, message_level)
 
-    async def _synchronize(self, head_level: int) -> None:
+    async def _synchronize(self, sync_level: int) -> None:
         """Fetch operations via Fetcher and pass to message callback"""
-        index_level = await self._enter_sync_state(head_level)
+        index_level = await self._enter_sync_state(sync_level)
         if index_level is None:
             return
 
@@ -690,15 +638,15 @@ class BigMapIndex(Index):
                 self._config.skip_history == SkipHistory.once and not self.state.level,
             )
         ):
-            await self._synchronize_level(head_level)
+            await self._synchronize_level(sync_level)
         else:
-            await self._synchronize_full(index_level, head_level)
+            await self._synchronize_full(index_level, sync_level)
 
-        await self._exit_sync_state(head_level)
+        await self._exit_sync_state(sync_level)
 
-    async def _synchronize_full(self, index_level: int, head_level: int) -> None:
+    async def _synchronize_full(self, index_level: int, sync_level: int) -> None:
         first_level = index_level + 1
-        self._logger.info('Fetching big map diffs from level %s to %s', first_level, head_level)
+        self._logger.info('Fetching big map diffs from level %s to %s', first_level, sync_level)
 
         big_map_addresses = self._get_big_map_addresses()
         big_map_paths = self._get_big_map_paths()
@@ -706,7 +654,7 @@ class BigMapIndex(Index):
         fetcher = BigMapFetcher(
             datasource=self._datasource,
             first_level=first_level,
-            last_level=head_level,
+            last_level=sync_level,
             big_map_addresses=big_map_addresses,
             big_map_paths=big_map_paths,
         )
@@ -714,9 +662,9 @@ class BigMapIndex(Index):
         async for level, big_maps in fetcher.fetch_big_maps_by_level():
             with ExitStack() as stack:
                 if Metrics.enabled:
-                    Metrics.set_levels_to_sync(self._config.name, head_level - level)
+                    Metrics.set_levels_to_sync(self._config.name, sync_level - level)
                     stack.enter_context(Metrics.measure_level_sync_duration())
-                await self._process_level_big_maps(big_maps)
+                await self._process_level_big_maps(big_maps, sync_level)
 
     async def _synchronize_level(self, head_level: int) -> None:
         # NOTE: Checking late because feature flags could be modified after loading config
@@ -733,7 +681,7 @@ class BigMapIndex(Index):
                         big_map_ids.add((int(contract_big_map['ptr']), address, path))
 
         # NOTE: Do not use `_process_level_big_maps` here; we want to maintain transaction manually.
-        async with in_global_transaction():
+        async with self._ctx._transactions.in_transaction(head_level, head_level, self.name):
             for big_map_id, address, path in big_map_ids:
                 async for big_map_keys in self._datasource.iter_big_map(big_map_id, head_level):
                     big_map_data = tuple(
@@ -758,25 +706,31 @@ class BigMapIndex(Index):
 
             await self.state.update_status(level=head_level)
 
-    async def _process_level_big_maps(self, big_maps: Tuple[BigMapData, ...]) -> None:
+    async def _process_level_big_maps(
+        self,
+        big_maps: Tuple[BigMapData, ...],
+        sync_level: int,
+    ) -> None:
         if not big_maps:
             return
+
         batch_level = self._extract_level(big_maps)
         index_level = self.state.level
-
-        # NOTE: le operator because single level rollbacks are not supported
         if batch_level <= index_level:
-            raise RuntimeError(f'Level of big map batch must be higher than index state level: {batch_level} <= {index_level}')
+            raise RuntimeError(f'Batch level is lower than index level: {batch_level} <= {index_level}')
 
         self._logger.debug('Processing big map diffs of level %s', batch_level)
         matched_handlers = await self._match_big_maps(big_maps)
 
+        if Metrics.enabled:
+            Metrics.set_index_handlers_matched(len(matched_handlers))
+
         # NOTE: We still need to bump index level but don't care if it will be done in existing transaction
         if not matched_handlers:
             await self.state.update_status(level=batch_level)
             return
 
-        async with in_global_transaction():
+        async with self._ctx._transactions.in_transaction(batch_level, sync_level, self.name):
             for handler_config, big_map_diff in matched_handlers:
                 await self._call_matched_handler(handler_config, big_map_diff)
             await self.state.update_status(level=batch_level)
@@ -796,24 +750,16 @@ class BigMapIndex(Index):
     ) -> BigMapDiff:
         """Prepare handler arguments, parse key and value. Schedule callback in executor."""
         self._logger.info('%s: `%s` handler matched!', matched_big_map.operation_id, handler_config.callback)
-        if not handler_config.parent:
-            raise ConfigInitializationException
 
         if matched_big_map.action.has_key:
-            key_type = handler_config.key_type_cls
-            try:
-                key = key_type.parse_obj(matched_big_map.key)
-            except ValidationError as e:
-                raise InvalidDataError(key_type, matched_big_map.key, matched_big_map) from e
+            type_ = handler_config.key_type_cls
+            key = parse_object(type_, matched_big_map.key) if type_ else None
         else:
             key = None
 
         if matched_big_map.action.has_value:
-            value_type = handler_config.value_type_cls
-            try:
-                value = value_type.parse_obj(matched_big_map.value)
-            except ValidationError as e:
-                raise InvalidDataError(value_type, matched_big_map.key, matched_big_map) from e
+            type_ = handler_config.value_type_cls
+            value = parse_object(type_, matched_big_map.value) if type_ else None
         else:
             value = None
 
@@ -841,7 +787,7 @@ class BigMapIndex(Index):
         if not handler_config.parent:
             raise ConfigInitializationException
 
-        await self._ctx.fire_handler(
+        await self._ctx._fire_handler(
             handler_config.callback,
             handler_config.parent.name,
             self.datasource,
@@ -901,11 +847,10 @@ class HeadIndex(Index):
 
             batch_level = head.level
             index_level = self.state.level
-
             if batch_level <= index_level:
-                raise RuntimeError(f'Level of head must be higher than index state level: {batch_level} <= {index_level}')
+                raise RuntimeError(f'Batch level is lower than index level: {batch_level} <= {index_level}')
 
-            async with in_global_transaction():
+            async with self._ctx._transactions.in_transaction(batch_level, message_level, self.name):
                 self._logger.debug('Processing head info of level %s', batch_level)
                 for handler_config in self._config.handlers:
                     await self._call_matched_handler(handler_config, head)
@@ -915,7 +860,7 @@ class HeadIndex(Index):
         if not handler_config.parent:
             raise ConfigInitializationException
 
-        await self._ctx.fire_handler(
+        await self._ctx._fire_handler(
             handler_config.callback,
             handler_config.parent.name,
             self.datasource,
@@ -941,59 +886,71 @@ class TokenTransferIndex(Index):
         if Metrics.enabled:
             Metrics.set_levels_to_realtime(self._config.name, len(self._queue))
 
-    async def _synchronize(self, last_level: int) -> None:
+    async def _synchronize(self, sync_level: int) -> None:
         """Fetch token transfers via Fetcher and pass to message callback"""
-        first_level = await self._enter_sync_state(last_level)
+        first_level = await self._enter_sync_state(sync_level)
         if first_level is None:
             return
 
-        self._logger.info('Fetching token transfers from level %s to %s', first_level, last_level)
+        self._logger.info('Fetching token transfers from level %s to %s', first_level, sync_level)
 
         fetcher = TokenTransferFetcher(
             datasource=self._datasource,
             first_level=first_level,
-            last_level=last_level,
+            last_level=sync_level,
         )
 
         async for level, token_transfers in fetcher.fetch_token_transfers_by_level():
             with ExitStack() as stack:
                 if Metrics.enabled:
-                    Metrics.set_levels_to_sync(self._config.name, last_level - level)
+                    Metrics.set_levels_to_sync(self._config.name, sync_level - level)
                     stack.enter_context(Metrics.measure_level_sync_duration())
-                await self._process_level_token_transfers(token_transfers)
+                await self._process_level_token_transfers(token_transfers, sync_level)
 
-        await self._exit_sync_state(last_level)
+        await self._exit_sync_state(sync_level)
 
-    async def _process_level_token_transfers(self, token_transfers: Iterable[TokenTransferData]) -> None:
+    async def _process_level_token_transfers(
+        self,
+        token_transfers: Tuple[TokenTransferData, ...],
+        sync_level: int,
+    ) -> None:
         if not token_transfers:
             return
-        level = self._extract_level(tuple(token_transfers))
 
+        # FIXME: Why is this needed?
+        batch_level = self._extract_level(token_transfers)
         if self.state.status == IndexStatus.SYNCING:
-            if level < self.state.level:
-                raise RuntimeError(f'Level of token transfer batch must be not lower than index state level: {level} < {self.state.level}')
+            if batch_level < self.state.level:
+                raise RuntimeError(
+                    f'Level of token transfer batch must be not lower than index state level: {batch_level} < {self.state.level}'
+                )
         else:
-            if level <= self.state.level:
-                raise RuntimeError(f'Level of token transfer batch must be higher than index state level: {level} <= {self.state.level}')
+            if batch_level <= self.state.level:
+                raise RuntimeError(
+                    f'Level of token transfer batch must be higher than index state level: {batch_level} <= {self.state.level}'
+                )
 
-        self._logger.debug('Processing token transfers of level %s', level)
+        self._logger.debug('Processing token transfers of level %s', batch_level)
         matched_handlers = await self._match_token_transfers(token_transfers)
 
+        if Metrics.enabled:
+            Metrics.set_index_handlers_matched(len(matched_handlers))
+
         # NOTE: We still need to bump index level but don't care if it will be done in existing transaction
         if not matched_handlers:
-            await self.state.update_status(level=level)
+            await self.state.update_status(level=batch_level)
             return
 
-        async with in_global_transaction():
+        async with self._ctx._transactions.in_transaction(batch_level, sync_level, self.name):
             for handler_config, big_map_diff in matched_handlers:
                 await self._call_matched_handler(handler_config, big_map_diff)
-            await self.state.update_status(level=level)
+            await self.state.update_status(level=batch_level)
 
     async def _call_matched_handler(self, handler_config: TokenTransferHandlerConfig, token_transfer: TokenTransferData) -> None:
         if not handler_config.parent:
             raise ConfigInitializationException
 
-        await self._ctx.fire_handler(
+        await self._ctx._fire_handler(
             handler_config.callback,
             handler_config.parent.name,
             self.datasource,
@@ -1026,10 +983,10 @@ class TokenTransferIndex(Index):
             with ExitStack() as stack:
                 if Metrics.enabled:
                     stack.enter_context(Metrics.measure_level_realtime_duration())
-                await self._process_level_token_transfers(token_transfers)
+                await self._process_level_token_transfers(token_transfers, message_level)
 
     def _get_token_addresses(self) -> Set[str]:
-        """Get addresses to fetch big map diffs from during initial synchronization"""
+        """Get addresses to fetch transfers from during initial synchronization"""
         addresses: set[str] = set()
         for handler_config in self._config.handlers:
             if isinstance(handler_config, TokenTransferHandlerConfig):
@@ -1040,7 +997,7 @@ class TokenTransferIndex(Index):
         return addresses
 
     def _get_token_ids(self) -> Set[int]:
-        """Get addresses to fetch big map diffs from during initial synchronization"""
+        """Get token ids to fetch transfers from during initial synchronization"""
         ids: set[int] = set()
         for handler_config in self._config.handlers:
             if isinstance(handler_config, TokenTransferHandlerConfig):
@@ -1050,3 +1007,184 @@ class TokenTransferIndex(Index):
             if handler_config.token_id is not None:
                 ids.add(handler_config.token_id)
         return ids
+
+
+class EventIndex(Index):
+    message_type = MessageType.event
+    _config: EventIndexConfig
+
+    def __init__(self, ctx: DipDupContext, config: EventIndexConfig, datasource: TzktDatasource) -> None:
+        super().__init__(ctx, config, datasource)
+        self._queue: Deque[Tuple[EventData, ...]] = deque()
+
+    def push_events(self, events: Tuple[EventData, ...]) -> None:
+        self._queue.append(events)
+
+        if Metrics.enabled:
+            Metrics.set_levels_to_realtime(self._config.name, len(self._queue))
+
+    async def _process_queue(self) -> None:
+        """Process WebSocket queue"""
+        if self._queue:
+            self._logger.debug('Processing websocket queue')
+        while self._queue:
+            events = self._queue.popleft()
+            message_level = events[0].level
+            if message_level <= self.state.level:
+                self._logger.debug('Skipping outdated message: %s <= %s', message_level, self.state.level)
+                continue
+
+            with ExitStack() as stack:
+                if Metrics.enabled:
+                    stack.enter_context(Metrics.measure_level_realtime_duration())
+                await self._process_level_events(events, message_level)
+
+    async def _synchronize(self, sync_level: int) -> None:
+        """Fetch operations via Fetcher and pass to message callback"""
+        index_level = await self._enter_sync_state(sync_level)
+        if index_level is None:
+            return
+
+        first_level = index_level + 1
+        self._logger.info('Fetching contract events from level %s to %s', first_level, sync_level)
+
+        event_addresses = self._get_event_addresses()
+        event_tags = self._get_event_tags()
+
+        fetcher = EventFetcher(
+            datasource=self._datasource,
+            first_level=first_level,
+            last_level=sync_level,
+            event_addresses=event_addresses,
+            event_tags=event_tags,
+        )
+
+        async for level, events in fetcher.fetch_events_by_level():
+            with ExitStack() as stack:
+                if Metrics.enabled:
+                    Metrics.set_levels_to_sync(self._config.name, sync_level - level)
+                    stack.enter_context(Metrics.measure_level_sync_duration())
+                await self._process_level_events(events, sync_level)
+
+        await self._exit_sync_state(sync_level)
+
+    async def _process_level_events(
+        self,
+        events: Tuple[EventData, ...],
+        sync_level: int,
+    ) -> None:
+        if not events:
+            return
+
+        batch_level = self._extract_level(events)
+        index_level = self.state.level
+        if batch_level <= index_level:
+            raise RuntimeError(f'Batch level is lower than index level: {batch_level} <= {index_level}')
+
+        self._logger.debug('Processing contract events of level %s', batch_level)
+        matched_handlers = await self._match_events(events)
+
+        if Metrics.enabled:
+            Metrics.set_index_handlers_matched(len(matched_handlers))
+
+        # NOTE: We still need to bump index level but don't care if it will be done in existing transaction
+        if not matched_handlers:
+            await self.state.update_status(level=batch_level)
+            return
+
+        async with self._ctx._transactions.in_transaction(batch_level, sync_level, self.name):
+            for handler_config, event in matched_handlers:
+                await self._call_matched_handler(handler_config, event)
+            await self.state.update_status(level=batch_level)
+
+    async def _match_event(self, handler_config: EventHandlerConfigT, event: EventData) -> bool:
+        """Match single contract event with pattern"""
+        if isinstance(handler_config, EventHandlerConfig) and handler_config.tag != event.tag:
+            return False
+        if handler_config.contract_config.address != event.contract_address:
+            return False
+        return True
+
+    async def _prepare_handler_args(
+        self,
+        handler_config: EventHandlerConfigT,
+        matched_event: EventData,
+    ) -> Event | UnknownEvent | None:
+        """Prepare handler arguments, parse key and value. Schedule callback in executor."""
+        self._logger.info('%s: `%s` handler matched!', matched_event.level, handler_config.callback)
+
+        if isinstance(handler_config, UnknownEventHandlerConfig):
+            return UnknownEvent(
+                data=matched_event,
+                payload=matched_event.payload,
+            )
+
+        with suppress(InvalidDataError):
+            type_ = handler_config.event_type_cls
+            payload = parse_object(type_, matched_event.payload)
+            return Event(
+                data=matched_event,
+                payload=payload,
+            )
+
+        return None
+
+    async def _match_events(self, events: Iterable[EventData]) -> Deque[MatchedEventsT]:
+        """Try to match contract events with all index handlers."""
+        matched_handlers: Deque[MatchedEventsT] = deque()
+        events = deque(events)
+
+        for handler_config in self._config.handlers:
+            # NOTE: Matched events are dropped after processing
+            for event in copy(events):
+                if not await self._match_event(handler_config, event):
+                    continue
+
+                arg = await self._prepare_handler_args(handler_config, event)
+                if isinstance(arg, Event) and isinstance(handler_config, EventHandlerConfig):
+                    matched_handlers.append((handler_config, arg))
+                elif isinstance(arg, UnknownEvent) and isinstance(handler_config, UnknownEventHandlerConfig):
+                    matched_handlers.append((handler_config, arg))
+                elif arg is None:
+                    continue
+                else:
+                    raise RuntimeError
+
+                events.remove(event)
+
+        # NOTE: We don't care about `merge_subscriptions` here implying that all events will be processed
+        # NOTE: Maybe "unfiltered" indexes will cover that case?
+        for address in {event.contract_address for event in events}:
+            self._logger.warning('Some events were not matched; fallback handler is missing for `{}`', address)
+
+        return matched_handlers
+
+    async def _call_matched_handler(self, handler_config: EventHandlerConfigT, event: Event | UnknownEvent):
+        if isinstance(handler_config, EventHandlerConfig) != isinstance(event, Event):
+            raise RuntimeError(f'Invalid handler config and event types: {handler_config}, {event}')
+
+        if not handler_config.parent:
+            raise ConfigInitializationException
+
+        await self._ctx._fire_handler(
+            handler_config.callback,
+            handler_config.parent.name,
+            self.datasource,
+            str(event.data.transaction_id),
+            event,
+        )
+
+    def _get_event_addresses(self) -> Set[str]:
+        """Get addresses to fetch events during initial synchronization"""
+        addresses = set()
+        for handler_config in self._config.handlers:
+            addresses.add(cast(ContractConfig, handler_config.contract).address)
+        return addresses
+
+    def _get_event_tags(self) -> Set[str]:
+        """Get tags to fetch events during initial synchronization"""
+        paths = set()
+        for handler_config in self._config.handlers:
+            if isinstance(handler_config, EventHandlerConfig):
+                paths.add(handler_config.tag)
+        return paths
