@@ -40,12 +40,13 @@ from dipdup.enums import IndexStatus
 from dipdup.enums import IndexType
 from dipdup.enums import ReindexingReason
 from dipdup.enums import TokenStandard
-from dipdup.utils import json_dumps
+from dipdup.utils import json_dumps_decimals
 
 ParameterType = TypeVar('ParameterType', bound=BaseModel)
 StorageType = TypeVar('StorageType', bound=BaseModel)
 KeyType = TypeVar('KeyType', bound=BaseModel)
 ValueType = TypeVar('ValueType', bound=BaseModel)
+EventType = TypeVar('EventType', bound=BaseModel)
 
 
 _logger = logging.getLogger(__name__)
@@ -234,6 +235,33 @@ class TokenTransferData:
     tzkt_migration_id: Optional[int] = None
 
 
+@dataclass
+class EventData:
+    """Basic structure for events received from TzKT REST API"""
+
+    id: int
+    level: int
+    timestamp: datetime
+    tag: str
+    payload: Any | None
+    contract_address: str
+    contract_alias: Optional[str] = None
+    contract_code_hash: Optional[int] = None
+    transaction_id: Optional[int] = None
+
+
+@dataclass
+class Event(Generic[EventType]):
+    data: EventData
+    payload: EventType
+
+
+@dataclass
+class UnknownEvent:
+    data: EventData
+    payload: Any | None
+
+
 # ===> Model Versioning
 
 
@@ -278,7 +306,7 @@ class ModelUpdate(TortoiseModel):
     index = fields.CharField(256)
 
     action = fields.CharEnumField(ModelUpdateAction)
-    data: Dict[str, Any] = fields.JSONField(encoder=json_dumps, null=True)
+    data: Dict[str, Any] = fields.JSONField(encoder=json_dumps_decimals, null=True)
 
     created_at = fields.DatetimeField(auto_now_add=True)
     updated_at = fields.DatetimeField(auto_now=True)
